import unittest
import sys
import os

from giscanner.shlibs import resolve_from_ldd_output, sanitize_shlib_path


class TestLddParser(unittest.TestCase):

    def test_resolve_from_ldd_output(self):
        output = '''\
            libglib-2.0.so.0 => /usr/lib/x86_64-linux-gnu/libglib-2.0.so.0 (0x00007fbe12d68000)
            libgtk-3.so.0 => /usr/lib/x86_64-linux-gnu/libgtk-3.so.0 (0x00007fbe12462000)
            libgdk-3.so.0 => /usr/lib/x86_64-linux-gnu/libgdk-3.so.0 (0x00007fbe1216c000)
            libpango-1.0.so.0 => /usr/lib/x86_64-linux-gnu/libpango-1.0.so.0 (0x00007fbe11d1a000)
            libatk-1.0.so.0 => /usr/lib/x86_64-linux-gnu/libatk-1.0.so.0 (0x00007fbe11af4000)'''
        libraries = ['glib-2.0', 'gtk-3', 'pango-1.0']

        self.assertEqual(
            ['libglib-2.0.so.0', 'libgtk-3.so.0', 'libpango-1.0.so.0'],
            resolve_from_ldd_output(libraries, output))

    @unittest.skipUnless(os.name == "posix", "posix only")
    def test_resolve_from_ldd_output_macos_rpath(self):
        output = '''\
            @rpath/libbarapp-1.0.dylib (compatibility version 0.0.0, current version 0.0.0)
            /foo/libgio-2.0.0.dylib (compatibility version 5801.0.0, current version 5801.3.0)
            /foo/libgmodule-2.0.0.dylib (compatibility version 5801.0.0, current version 5801.3.0)'''

        libraries = ['barapp-1.0']
        shlibs = resolve_from_ldd_output(libraries, output)
        self.assertEqual(shlibs, ['@rpath/libbarapp-1.0.dylib'])
        self.assertEqual(sanitize_shlib_path(shlibs[0]), 'libbarapp-1.0.dylib')

    @unittest.skipUnless(os.name == "posix", "posix only")
    def test_sanitize_shlib_path(self):
        self.assertEqual(
            sanitize_shlib_path('@rpath/libbarapp-1.0.dylib'),
            'libbarapp-1.0.dylib')

        self.assertEqual(
            sanitize_shlib_path('/foo/bar'),
            '/foo/bar' if sys.platform == 'darwin' else 'bar')

    def test_unresolved_library(self):
        output = ''
        libraries = ['foo']

        with self.assertRaises(SystemExit, msg='can\'t resolve libraries to shared libraries: foo'):
            resolve_from_ldd_output(libraries, output)

    def test_prefixed_library_name(self):
        output = '''\
           /usr/lib/liblibX.so
           /usr/lib/libX.so'''

        self.assertEqual(
            ['/usr/lib/libX.so'],
            resolve_from_ldd_output(['X'], output))
        self.assertEqual(
            ['/usr/lib/liblibX.so'],
            resolve_from_ldd_output(['libX'], output))

    def test_suffixed_library_name(self):
        output = '''\
            libpangocairo.so.0 => /usr/lib/x86_64-linux-gnu/libpangocairo.so.0 (0x00)
            libpangoft2.so.0 => /usr/lib/x86_64-linux-gnu/libpangoft2.so.0 (0x00)
            libpango.so.0 => /usr/lib/x86_64-linux-gnu/libpango.so.0 (0x00)'''
        libraries = ['pango']

        self.assertEqual(
            ['libpango.so.0'],
            resolve_from_ldd_output(libraries, output))

    def test_header_is_ignored(self):
        output = '''/tmp-introspection/libfoo.so.999:
            0000000000000000 0000000000000000 rlib  0    3   0      /usr/local/lib/libfoo.so.1'''
        libraries = ['foo']

        self.assertEqual(
            ['/usr/local/lib/libfoo.so.1'],
            resolve_from_ldd_output(libraries, output))

    def test_executable_path_includes_library_name(self):
        """
        Regression test for https://gitlab.gnome.org/GNOME/gobject-introspection/issues/208
        """
        output = '''/usr/ports/pobj/libgepub-0.6.0/build-amd64/tmp-introspectnxmyodg1/Gepub-0.6:
            Start            End              Type  Open Ref GrpRef Name
            00001066c8400000 00001066c8605000 exe   2    0   0      /usr/ports/pobj/libgepub-0.6.0/build-amd64/tmp-introspectnxmyodg1/Gepub-0.6
            000010690019c000 00001069003a8000 rlib  0    1   0      /usr/local/lib/libgepub-0.6.so.0.0'''
        libraries = ['gepub-0.6']

        self.assertEqual(
            ['/usr/local/lib/libgepub-0.6.so.0.0'],
            resolve_from_ldd_output(libraries, output))

    def test_library_path_includes_library_name(self):
        output = '''/usr/ports/pobj/gnome-music-3.28.1/build-amd64/tmp-introspectuz5xaun3/Gd-1.0:
            Start            End              Type  Open Ref GrpRef Name
            0000070e40f00000 0000070e41105000 exe   2    0   0      /usr/ports/pobj/gnome-music-3.28.1/build-amd64/tmp-introspectuz5xaun3/Gd-1.0
            00000710f9b39000 00000710f9d51000 rlib  0    1   0      /usr/ports/pobj/gnome-music-3.28.1/build-amd64/subprojects/libgd/libgd/libgd.so'''
        libraries = ['gd']

        self.assertEqual(
            ['/usr/ports/pobj/gnome-music-3.28.1/build-amd64/subprojects/libgd/libgd/libgd.so'],
            resolve_from_ldd_output(libraries, output))

    def test_basename(self):
        output = '''/usr/lib/libfoo.so'''

        self.assertEqual(
            ['/usr/lib/libfoo.so'],
            resolve_from_ldd_output(['foo'], output))


if __name__ == '__main__':
    unittest.main()



import unittest
import tempfile
import os
import sys
import textwrap


os.environ['GI_SCANNER_DISABLE_CACHE'] = '1'

from giscanner import ast
from giscanner.sourcescanner import SourceScanner
from giscanner.transformer import Transformer
from giscanner.message import MessageLogger


def create_scanner_from_source_string(source):
    ss = SourceScanner()
    tmp_fd, tmp_name = tempfile.mkstemp(suffix='.h', text=True)

    try:
        with os.fdopen(tmp_fd, 'w', encoding='utf-8') as file:
            file.write(source)
        ss.parse_files([tmp_name])
    finally:
        os.unlink(tmp_name)

    return ss


def load_namespace_from_source_string(namespace, source):
    ss = create_scanner_from_source_string(source)
    xformer = Transformer(namespace)
    xformer.parse(ss.get_symbols())


class TestIdentifierFilter(unittest.TestCase):
    def test_underscore_t_identifier_filter(self):
        cmd = [sys.executable, '-c', textwrap.dedent("""
            import sys, re
            for line in sys.stdin:
                line = re.sub(r"^test_t$", "TestContext", line)
                line = re.sub(r"_t$", "", line)
                line = re.sub(r"^test_", "Test_", line)
                line = re.sub(r"_([a-z])", lambda m: m.group(1).title(),
                              line)
                sys.stdout.write(line)""")]

        namespace = ast.Namespace('Test', '1.0')
        xformer = Transformer(namespace, identifier_filter_cmd=cmd)

        self.assertEqual(xformer.strip_identifier('test_t'), 'Context')
        self.assertEqual(xformer.strip_identifier('test_foo_t'), 'Foo')
        self.assertEqual(xformer.strip_identifier('test_foot'), 'Foot')
        self.assertEqual(xformer.strip_identifier('test_foo_bart'), 'FooBart')
        self.assertEqual(xformer.strip_identifier('test_foo_tart'), 'FooTart')

    def test_invalid_command(self):
        cmd = ['this-is-not-a-real-command']
        namespace = ast.Namespace('Test', '1.0')
        xformer = Transformer(namespace, identifier_filter_cmd=cmd)
        self.assertRaises(OSError, xformer.strip_identifier, 'test_t')

    def test_invalid_argument(self):
        cmd = [sys.executable, '--not-a-valid-argument']
        namespace = ast.Namespace('Test', '1.0')
        xformer = Transformer(namespace, identifier_filter_cmd=cmd)
        self.assertRaises(ValueError, xformer.strip_identifier, 'test_t')


class TestSymbolFilter(unittest.TestCase):

    def test_split_csymbol(self):
        cmd = [
            sys.executable, '-c',
            'import sys; sys.stdout.write("test_" + sys.stdin.read())']
        namespace = ast.Namespace('Test', '1.0')
        xformer = Transformer(namespace, symbol_filter_cmd=cmd)

        self.assertEqual(
            xformer.split_csymbol('foo_bar_quux')[1], "foo_bar_quux")


class TestStructTypedefs(unittest.TestCase):
    def setUp(self):
        # Hack to set logging singleton
        self.namespace = ast.Namespace('Test', '1.0')
        logger = MessageLogger.get(namespace=self.namespace)
        logger.enable_warnings(True)

    def test_anonymous_typedef(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                int value;
            } TestStruct;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)

    def test_typedef_before(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct _TestStruct TestStruct;
            struct _TestStruct {
                int value;
            };
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)

    def test_typedef_after(self):
        load_namespace_from_source_string(self.namespace, """
            struct _TestStruct {
                int value;
            };
            typedef struct _TestStruct TestStruct;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)

    def test_tag_and_typedef(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct _TestStruct {
                int value;
            } TestStruct;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)

    def test_struct_tag_only(self):
        load_namespace_from_source_string(self.namespace, """
            struct TestStruct {
                int value;
            };
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)

    def test_struct_aliases(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                int value;
            } TestStruct;
            typedef TestStruct TestStructAlias1;
            typedef TestStruct TestStructAlias2;
            """)
        self.assertEqual(len(self.namespace.names), 3)

        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)

        alias = self.namespace.get('StructAlias1')
        self.assertIsInstance(alias, ast.Alias)
        self.assertEqual(alias.target, node)
        self.assertEqual(alias.ctype, 'TestStructAlias1')

        alias = self.namespace.get('StructAlias2')
        self.assertIsInstance(alias, ast.Alias)
        self.assertEqual(alias.target, node)
        self.assertEqual(alias.ctype, 'TestStructAlias2')

    def test_struct_tag_aliases_before(self):
        # This test is similar to how GObject and GInitiallyUnowned are setup
        load_namespace_from_source_string(self.namespace, """
            typedef struct _TestStruct TestStruct;
            typedef struct _TestStruct TestStructAlias;
            struct _TestStruct {
                int value;
            };
            """)
        self.assertEqual(len(self.namespace.names), 2)

        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)
        self.assertEqual(node.ctype, 'TestStruct')

        shared = self.namespace.get('StructAlias')
        self.assertIsNotNone(shared)
        self.assertIsInstance(shared, ast.Record)
        self.assertFalse(shared.opaque)
        self.assertFalse(shared.pointer)
        self.assertFalse(shared.disguised)
        self.assertEqual(len(shared.fields), 1)
        self.assertEqual(shared.ctype, 'TestStructAlias')

    def test_struct_tag_aliases_after(self):
        load_namespace_from_source_string(self.namespace, """
            struct _TestStruct {
                int value;
            };
            typedef struct _TestStruct TestStruct;
            typedef struct _TestStruct TestStructAlias;
            """)
        self.assertEqual(len(self.namespace.names), 2)

        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(len(node.fields), 1)
        self.assertEqual(node.ctype, 'TestStruct')

        shared = self.namespace.get('StructAlias')
        self.assertIsNotNone(shared)
        self.assertIsInstance(shared, ast.Record)
        self.assertFalse(shared.opaque)
        self.assertFalse(shared.pointer)
        self.assertFalse(shared.disguised)
        self.assertEqual(len(shared.fields), 1)
        self.assertEqual(shared.ctype, 'TestStructAlias')

    def test_struct_pointer(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                int value;
            } TestStruct;
            typedef TestStruct* TestStructPtr;
            """)
        self.assertEqual(len(self.namespace.names), 2)

        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertEqual(len(node.fields), 1)
        self.assertEqual(node.ctype, 'TestStruct')

        ptr = self.namespace.get('StructPtr')
        self.assertIsNotNone(ptr)
        self.assertIsInstance(ptr, ast.Alias)
        self.assertEqual(ptr.ctype, 'TestStructPtr')

    def test_struct_tag_pointer(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct _TestStruct TestStruct;
            struct _TestStruct{
                int value;
            };
            typedef struct _TestStruct* TestStructPtr;
            """)
        self.assertEqual(len(self.namespace.names), 2)

        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertFalse(node.opaque)
        self.assertFalse(node.pointer)
        self.assertFalse(node.disguised)
        self.assertEqual(node.ctype, 'TestStruct')
        self.assertEqual(len(node.fields), 1)

        ptr = self.namespace.get('StructPtr')
        self.assertIsNotNone(ptr)
        # This currently gives a disguised Record instead of an Alias
        self.assertIsInstance(ptr, ast.Record)
        self.assertFalse(ptr.opaque)
        self.assertTrue(ptr.pointer)
        self.assertTrue(ptr.disguised)
        self.assertEqual(len(ptr.fields), 0)
        self.assertEqual(ptr.ctype, 'TestStructPtr')

    def test_opaque_struct(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct _TestStruct TestStruct;
            """)
        self.assertEqual(len(self.namespace.names), 1)

        node = self.namespace.get('Struct')
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ast.Record)
        self.assertTrue(node.opaque)
        self.assertFalse(node.pointer)
        self.assertTrue(node.disguised)
        self.assertEqual(node.ctype, 'TestStruct')
        self.assertFalse(node.fields)


class TestNestedStructs(unittest.TestCase):
    def setUp(self):
        # Hack to set logging singleton
        self.namespace = ast.Namespace('Test', '1.0')
        logger = MessageLogger.get(namespace=self.namespace)
        logger.enable_warnings(True)

    def test_nested_named(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                int value;
            } TestSimpleStruct;

            typedef struct {
                TestSimpleStruct nested_struct;
            } TestStructWithNested;
            """)
        self.assertEqual(len(self.namespace.names), 2)
        node = self.namespace.get('StructWithNested')
        self.assertEqual(len(node.fields), 1)

        simple = self.namespace.get('SimpleStruct')
        self.assertIsNotNone(node)

        field = node.fields[0]
        self.assertIsNotNone(field)
        self.assertIsInstance(field, ast.Field)
        self.assertEqual(field.type, simple)
        self.assertEqual(field.name, 'nested_struct')

    def test_nested_anonymous(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                struct {
                    int value;
                };
            } TestStructWithNestedAnon;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('StructWithNestedAnon')
        self.assertEqual(len(node.fields), 1)

        field = node.fields[0]
        self.assertIsNotNone(field)
        self.assertIsInstance(field, ast.Field)
        self.assertIsNone(field.name)

        anon = field.anonymous_node
        self.assertIsNotNone(anon)
        self.assertIsInstance(anon, ast.Record)
        self.assertEqual(len(anon.fields), 1)

        anon_field = anon.fields[0]
        self.assertIsNotNone(anon_field)
        self.assertIsInstance(anon_field, ast.Field)
        self.assertEqual(anon_field.name, 'value')

    def test_nested(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                struct {
                    int value;
                } nested;
            } TestStructWithNested;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('StructWithNested')
        self.assertEqual(len(node.fields), 1)

        field = node.fields[0]
        self.assertIsNotNone(field)
        self.assertIsInstance(field, ast.Field)
        self.assertEqual(field.name, 'nested')

        nested = field.anonymous_node
        self.assertIsNotNone(nested)
        self.assertIsInstance(nested, ast.Record)
        self.assertEqual(len(nested.fields), 1)
        self.assertEqual(nested.name, 'nested')

        nested_field = nested.fields[0]
        self.assertIsNotNone(nested_field)
        self.assertIsInstance(nested_field, ast.Field)
        self.assertEqual(nested_field.name, 'value')

    def test_struct_ptr(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                int value;
            } TestSimpleStruct;

            typedef struct {
                TestSimpleStruct *struct_ptr;
            } TestStructWithNestedPtr;
            """)
        self.assertEqual(len(self.namespace.names), 2)
        node = self.namespace.get('StructWithNestedPtr')
        self.assertEqual(len(node.fields), 1)

        simple = self.namespace.get('SimpleStruct')
        self.assertIsNotNone(simple)

        field = node.fields[0]
        self.assertIsNotNone(field)
        self.assertIsInstance(field, ast.Field)
        self.assertEqual(field.type.ctype, 'TestSimpleStruct*')
        self.assertEqual(field.name, 'struct_ptr')


class TestUnions(unittest.TestCase):
    def setUp(self):
        # Hack to set logging singleton
        self.namespace = ast.Namespace('Test', '1.0')
        logger = MessageLogger.get(namespace=self.namespace)
        logger.enable_warnings(True)

    def test_union_with_struct(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                int value;
            } TestSimpleStruct;

            typedef union {
                TestSimpleStruct nested_struct;
                int value;
            } TestUnionWithNested;
            """)
        self.assertEqual(len(self.namespace.names), 2)
        node = self.namespace.get('UnionWithNested')
        self.assertEqual(len(node.fields), 2)

        simple = self.namespace.get('SimpleStruct')
        self.assertIsNotNone(node)

        field = node.fields[0]
        self.assertEqual(field.type, simple)
        self.assertEqual(field.name, 'nested_struct')

        field = node.fields[1]
        self.assertEqual(field.type.ctype, 'int')
        self.assertEqual(field.name, 'value')

    def test_nested(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                union {
                    int ival;
                    float fval;
                } nested;
            } TestNestedUnion;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('NestedUnion')
        self.assertEqual(len(node.fields), 1)

        field = node.fields[0]
        self.assertEqual(field.name, 'nested')

        nested = field.anonymous_node
        self.assertIsInstance(nested, ast.Union)
        self.assertEqual(nested.name, 'nested')
        self.assertEqual(len(nested.fields), 2)

        self.assertEqual(nested.fields[0].name, 'ival')
        self.assertEqual(nested.fields[1].name, 'fval')

    def test_nested_anonymous(self):
        load_namespace_from_source_string(self.namespace, """
            typedef struct {
                union {
                    int ival;
                    float fval;
                };
            } TestStructWithNestedUnion;
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('StructWithNestedUnion')
        self.assertEqual(len(node.fields), 1)

        field = node.fields[0]
        self.assertIsNone(field.name)

        nested = field.anonymous_node
        self.assertIsInstance(nested, ast.Union)
        self.assertIsNone(nested.name)
        self.assertEqual(len(nested.fields), 2)

        self.assertEqual(nested.fields[0].name, 'ival')
        self.assertEqual(nested.fields[0].type.ctype, 'int')

        self.assertEqual(nested.fields[1].name, 'fval')
        self.assertEqual(nested.fields[1].type.ctype, 'float')


class TestCallbacks(unittest.TestCase):
    def setUp(self):
        # Hack to set logging singleton
        self.namespace = ast.Namespace('Test', '1.0')
        logger = MessageLogger.get(namespace=self.namespace)
        logger.enable_warnings(True)

    def test_union_with_struct(self):
        load_namespace_from_source_string(self.namespace, """
            typedef void (*TestCallback)(int value);
            """)
        self.assertEqual(len(self.namespace.names), 1)
        node = self.namespace.get('Callback')
        self.assertIsInstance(node, ast.Callback)


class TestArrays(unittest.TestCase):
    def setUp(self):
        # Hack to set logging singleton
        self.namespace = ast.Namespace('Test', '1.0')
        logger = MessageLogger.get(namespace=self.namespace)
        logger.enable_warnings(True)

    def test_multidimensional_arrays(self):
        """Multidimensional arrays are flattend into a single dimension."""

        load_namespace_from_source_string(self.namespace, """
            typedef struct {
              int data[2][3][5][7][11];
            } TestMultiDimArray;
            """)

        node = self.namespace.get('MultiDimArray')
        self.assertIsNotNone(node)

        field = node.fields[0]
        self.assertIsInstance(field, ast.Field)
        self.assertIsInstance(field.type, ast.Array)
        self.assertEqual(field.type.element_type, ast.TYPE_INT)
        self.assertEqual(field.type.size, 2 * 3 * 5 * 7 * 11)

    def test_flexible_array_member(self):
        """Flexible array members don't break transformer.

        They are generally unsupported, so nothing else is validated.
        """

        load_namespace_from_source_string(self.namespace, """
            typedef struct {
              int length;
              unsigned char data[];
            } TestFlexibleArray;
            """)

        node = self.namespace.get('FlexibleArray')
        self.assertIsNotNone(node)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Example invocation
# cd /path/to/gobject-introspection
# python3 misc/update-vulkan-gir.py -i /usr/include/vulkan/vulkan.h -o gir/Vulkan-1.0.gir --extra-cpp-args='-I/path/to/pycparser/utils/fake_libc_include/'

import sys
import argparse

from pycparser import c_parser, c_ast, parse_file

class TypedefVisitor(c_ast.NodeVisitor):
    def __init__ (self, f):
        super().__init__()
        self.f = f

    def visit_Typedef(self, node):
        if node.name.startswith("Vk"):
            self.f.write("<record name=\"" + str(node.name)[2:] + "\" c:type=\"" + str(node.name) + "\"/>\n")

def show_func_defs(filename, output, extra_cpp_args=None):
    # Note that cpp is used. Provide a path to your own cpp or
    # make sure one exists in PATH.
    ast = parse_file(filename, use_cpp=True,
                     cpp_args=extra_cpp_args)

    with open(output, 'w', encoding='utf-8') as f:
        f.write("""<?xml version="1.0"?>
<repository version="1.2"
            xmlns="http://www.gtk.org/introspection/core/1.0"
            xmlns:c="http://www.gtk.org/introspection/c/1.0">
  <namespace name="Vulkan" version="1.0"
	     c:identifier-prefixes="VK"
	     c:symbol-prefixes="vk">
""")
        v = TypedefVisitor(f)
        v.visit(ast)
        f.write("""
  </namespace>
</repository>
""")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Parse Vulkan header files to produce Vulkan-1.0.gir")
    parser.add_argument('-i', '--input', type=str,
        help='The input vulkan header to process')
    parser.add_argument('--extra-cpp-args', type=str, default=None,
        help='Extra arguments to parse to the C Preprocesser.  e.g. the location pycparser\'s fake libc includes')
    parser.add_argument('-o', '--output', type=str, default='Vulkan-1.0.gir',
        help='The output Vulkan-1.0.gir location')
    args = parser.parse_args(sys.argv[1:])
    show_func_defs(args.input, args.output, args.extra_cpp_args)



import argparse
import os
import sys

try:
    from lxml import etree
except ImportError:
    # Skip the test if lxml is not available
    sys.exit(77)


args = argparse.ArgumentParser(prog="validate-gir")
args.add_argument("-S", "--schema", metavar="FILE", help="the schema file", required=True)
args.add_argument("infile", metavar="GIRFILE", type=argparse.FileType('r', encoding='UTF-8'),
                  default=sys.stdin, help="the GIR file to validate")

options = args.parse_args(sys.argv[1:])

with open(options.schema, "r", encoding="utf-8") as f:
    schema = etree.RelaxNG.from_rnc_string(f.read())

gir = etree.parse(options.infile)
if not schema.validate(gir):
    print(f"Validation errors: {schema.error_log}", file=sys.stderr)
    sys.exit(1)



# -*- Mode: Python -*-
# GObject-Introspection - a framework for introspecting GObject libraries
# Copyright (C) 2008 Colin Walters
# Copyright (C) 2008 Johan Dahlin
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

import os
import sys
import shlex
import subprocess
import tempfile

from .gdumpparser import IntrospectionBinary
from . import pkgconfig, utils
from .ccompiler import CCompiler

# bugzilla.gnome.org/558436
# Compile a binary program which is then linked to a library
# we want to introspect, in order to call its get_type functions.

_PROGRAM_TEMPLATE = """/* This file is generated, do not edit */

#undef GLIB_VERSION_MIN_REQUIRED
#undef GLIB_VERSION_MAX_ALLOWED

#include <glib.h>
#include <string.h>
#include <stdlib.h>

%(gdump_include)s

int
main(int argc, char **argv)
{
  GError *error = NULL;
  const char *introspect_dump_prefix = "--introspect-dump=";

#if !GLIB_CHECK_VERSION(2,35,0)
  g_type_init ();
#endif

  %(init_sections)s

  if (argc != 2 || !g_str_has_prefix (argv[1], introspect_dump_prefix))
    {
      g_printerr ("Usage: %%s --introspect-dump=input,output", argv[0]);
      exit (1);
    }

  if (!dump_irepository (argv[1] + strlen(introspect_dump_prefix), &error))
    {
      g_printerr ("%%s\\n", error->message);
      exit (1);
    }
  exit (0);
}
"""


class CompilerError(Exception):
    pass


class LinkerError(Exception):
    pass


class DumpCompiler(object):

    _compiler = None

    def __init__(self, options, get_type_functions, error_quark_functions):
        self._options = options
        self._get_type_functions = get_type_functions
        self._error_quark_functions = error_quark_functions

        # Acquire the compiler (and linker) commands via the CCompiler class in ccompiler.py
        self._compiler = CCompiler()

        self._uninst_srcdir = os.environ.get('UNINSTALLED_INTROSPECTION_SRCDIR')
        self._packages = ['gio-2.0', 'gmodule-2.0']
        self._packages.extend(options.packages)

    # Public API

    def run(self):
        # We have to use the current directory to work around Unix
        # sysadmins who mount /tmp noexec
        tmpdir = tempfile.mkdtemp('', 'tmp-introspect', dir=os.getcwd())
        os.mkdir(os.path.join(tmpdir, '.libs'))

        tpl_args = {}
        if self._uninst_srcdir is not None:
            gdump_path = os.path.join(self._uninst_srcdir, 'giscanner', 'gdump.c')
        else:
            try:
                gdump_path = GDUMP_PATH
            except NameError:
                gdump_path = os.path.join(os.path.join(DATADIR),
                                          'gobject-introspection-1.0', 'gdump.c')
        if not os.path.isfile(gdump_path):
            raise SystemExit("Couldn't find %r" % (gdump_path, ))
        with open(gdump_path, encoding='utf-8') as gdump_file:
            gdump_contents = gdump_file.read()
        tpl_args['gdump_include'] = gdump_contents
        tpl_args['init_sections'] = "\n".join(self._options.init_sections)

        c_path = self._generate_tempfile(tmpdir, '.c')
        with open(c_path, 'w', encoding='utf-8') as f:
            f.write(_PROGRAM_TEMPLATE % tpl_args)

            # We need to reference our get_type and error_quark functions
            # to make sure they are pulled in at the linking stage if the
            # library is a static library rather than a shared library.
            if len(self._get_type_functions) > 0:
                for func in self._get_type_functions:
                    f.write("extern GType " + func + "(void);\n")
                f.write("G_MODULE_EXPORT GType (*GI_GET_TYPE_FUNCS_[])(void) = {\n")
                first = True
                for func in self._get_type_functions:
                    if first:
                        first = False
                    else:
                        f.write(",\n")
                    f.write("  " + func)
                f.write("\n};\n")
            if len(self._error_quark_functions) > 0:
                for func in self._error_quark_functions:
                    f.write("extern GQuark " + func + "(void);\n")
                f.write("G_MODULE_EXPORT GQuark (*GI_ERROR_QUARK_FUNCS_[])(void) = {\n")
                first = True
                for func in self._error_quark_functions:
                    if first:
                        first = False
                    else:
                        f.write(",\n")
                    f.write("  " + func)
                f.write("\n};\n")

        if self._compiler.compiler.exe_extension:
            ext = self._compiler.compiler.exe_extension
        else:
            ext = ''

        bin_path = self._generate_tempfile(tmpdir, ext)

        try:
            introspection_obj = self._compile(c_path)
        except CompilerError as e:
            if not utils.have_debug_flag('save-temps'):
                utils.rmtree(tmpdir)
            raise SystemExit('compilation of temporary binary failed:' + str(e))

        try:
            self._link(bin_path, introspection_obj)
        except LinkerError as e:
            if not utils.have_debug_flag('save-temps'):
                utils.rmtree(tmpdir)
            raise SystemExit('linking of temporary binary failed: ' + str(e))

        return IntrospectionBinary([bin_path], tmpdir)

    # Private API

    def _generate_tempfile(self, tmpdir, suffix=''):
        tmpl = '%s-%s%s' % (self._options.namespace_name,
                            self._options.namespace_version, suffix)
        return os.path.join(tmpdir, tmpl)

    def _compile(self, *sources):
        cflags = pkgconfig.cflags(self._packages,
                                  msvc_syntax=self._compiler.check_is_msvc())
        cflags.extend(self._options.cflags)
        return self._compiler.compile(cflags,
                                      self._options.cpp_includes,
                                      sources,
                                      self._options.init_sections)

    def _link(self, output, sources):
        args = []
        libtool = utils.get_libtool_command(self._options)
        if libtool:
            # Note: MSVC Builds do not use libtool!
            args.extend(libtool)
            args.append('--mode=link')
            args.append('--tag=CC')
            if self._options.quiet:
                args.append('--silent')

        args.extend(self._compiler.linker_cmd)

        # We can use -o for the Microsoft compiler/linker,
        # but it is considered deprecated usage
        if self._compiler.check_is_msvc():
            args.extend(['-out:' + output])
        else:
            args.extend(['-o', output])
            if libtool:
                if utils.host_os() == 'nt':
                    args.append('-Wl,--export-all-symbols')
                else:
                    args.append('-export-dynamic')

        if not self._compiler.check_is_msvc():
            # These envvars are not used for MSVC Builds!
            # MSVC Builds use the INCLUDE, LIB envvars,
            # which are automatically picked up during
            # compilation and linking
            for cppflag in shlex.split(os.environ.get('CPPFLAGS', '')):
                args.append(cppflag)
            for cflag in shlex.split(os.environ.get('CFLAGS', '')):
                args.append(cflag)

        # Make sure to list the library to be introspected first since it's
        # likely to be uninstalled yet and we want the uninstalled RPATHs have
        # priority (or we might run with installed library that is older)
        for source in sources:
            if not os.path.exists(source):
                raise CompilerError(
                    "Could not find object file: %s" % (source, ))

        args.extend(sources)

        pkg_config_libs = pkgconfig.libs(self._packages,
                                         msvc_syntax=self._compiler.check_is_msvc())

        if not self._options.external_library:
            self._compiler.get_internal_link_flags(args,
                                                   libtool,
                                                   self._options.libraries,
                                                   self._options.extra_libraries,
                                                   self._options.library_paths,
                                                   self._options.lib_dirs_envvar)
            args.extend(pkg_config_libs)

        else:
            args.extend(pkg_config_libs)
            self._compiler.get_external_link_flags(args, self._options.libraries)

        if not self._compiler.check_is_msvc():
            for ldflag in shlex.split(os.environ.get('LDFLAGS', '')):
                if ldflag != '-Wl,--as-needed':
                    args.append(ldflag)

        dll_dirs = utils.dll_dirs()
        dll_dirs.add_dll_dirs(self._packages)

        if not self._options.quiet:
            print("g-ir-scanner: link: %s" % (
                subprocess.list2cmdline(args), ))
            sys.stdout.flush()

        msys = os.environ.get('MSYSTEM', None)
        if msys:
            shell = os.environ.get('SHELL', 'sh.exe')
            # Create a temporary script file that
            # runs the command we want
            tf, tf_name = tempfile.mkstemp()
            with os.fdopen(tf, 'wb') as f:
                shellcontents = ' '.join([x.replace('\\', '/') for x in args])
                fcontents = '#!/bin/sh\nunset PWD\n{}\n'.format(shellcontents)
                if not isinstance(fcontents, bytes):
                    fcontents = fcontents.encode('utf-8')
                f.write(fcontents)
            shell = utils.which(shell)
            args = [shell, tf_name.replace('\\', '/')]
        try:
            subprocess.check_call(args)
        except subprocess.CalledProcessError as e:
            raise LinkerError(e)
        finally:
            if msys:
                os.remove(tf_name)
            dll_dirs.cleanup_dll_dirs()


def compile_introspection_binary(options, get_type_functions,
                                 error_quark_functions):
    dc = DumpCompiler(options, get_type_functions, error_quark_functions)
    return dc.run()



# -*- coding: utf-8 -*-

extensions = [
    'sphinx.ext.extlinks',
]
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['_build']

html_theme = 'sphinx_rtd_theme'
html_show_copyright = False
project = "GObject Introspection"
html_title = project

html_theme_options = {
    "display_version": False,
}

html_static_path = [
    "extra.css",
]

html_context = {
    'extra_css_files': [
        '_static/extra.css',
    ],
}

extlinks = {
    'bzbug': ('https://bugzilla.gnome.org/show_bug.cgi?id=%s', 'bz#'),
    'commit': ('https://gitlab.gnome.org/GNOME/gobject-introspection/commit/%s', ''),
    'issue': ('https://gitlab.gnome.org/GNOME/gobject-introspection/issues/%s', '#'),
    'mr': (
        'https://gitlab.gnome.org/GNOME/gobject-introspection/merge_requests/%s', '!'),
    'user': ('https://gitlab.gnome.org/%s', ''),
}



import unittest
import optparse
import os

from giscanner.scannermain import get_source_root_dirs


class TestScanner(unittest.TestCase):

    def test_get_source_root_dirs_options(self):
        options = optparse.Values({"sources_top_dirs": ["foo"]})
        paths = get_source_root_dirs(options, ["nope"])
        self.assertEqual(len(paths), 1)
        self.assertTrue(os.path.isabs(paths[0]))
        self.assertEqual(os.path.normcase(paths[0]),
                         os.path.normcase(os.path.join(os.getcwd(), "foo")))

    def test_get_source_root_dirs_guess(self):
        options = optparse.Values({"sources_top_dirs": []})
        cwd = os.getcwd()
        paths = get_source_root_dirs(
            options, [os.path.join(cwd, "foo"), os.path.join(cwd, "bar")])
        self.assertEqual(len(paths), 1)
        self.assertEqual(os.path.normcase(paths[0]), os.path.normcase(cwd))

        paths = get_source_root_dirs(options, [])
        self.assertEqual(paths, [])

    @unittest.skipUnless(os.name == "nt", "Windows only")
    def test_get_source_root_dirs_different_drives(self):
        options = optparse.Values({"sources_top_dirs": []})
        names = [
            os.path.join("X:", os.sep, "foo"),
            os.path.join("Y:", os.sep, "bar"),
        ]
        paths = get_source_root_dirs(options, names)
        self.assertEqual(paths, list(map(os.path.dirname, names)))


if __name__ == '__main__':
    unittest.main()



import os
import sys
import difflib
import builtins

path = os.getenv('UNINSTALLED_INTROSPECTION_SRCDIR', None)
assert path is not None
# Not correct, but enough to get the tests going uninstalled
builtins.__dict__['DATADIR'] = path
builtins.__dict__['GIR_DIR'] = path

path = os.getenv('TOP_BUILDDIR', None)
assert path is not None
sys.path.insert(0, path)

from giscanner.annotationparser import GtkDocCommentBlockParser
from giscanner.ast import Include, Namespace
from giscanner.introspectablepass import IntrospectablePass
from giscanner.maintransformer import MainTransformer
from giscanner.message import MessageLogger
from giscanner.sourcescanner import SourceScanner
from giscanner.transformer import Transformer
from giscanner.scannermain import process_packages


currentdir = os.path.dirname(os.path.abspath(sys.argv[0]))
current_name = os.path.basename(currentdir)
path = os.path.abspath(os.path.join(currentdir, '..', ''))
top_srcdir = os.environ['UNINSTALLED_INTROSPECTION_SRCDIR']
top_builddir = os.environ['TOP_BUILDDIR']


class ChunkedIO(object):
    def __init__(self):
        self.buffer = []

    def write(self, s):
        self.buffer.append(s)

    def getvalue(self):
        return self.buffer


class Options:
    def __init__(self):
        self.cpp_includes = []
        self.cpp_defines = []
        self.cpp_undefines = []
        self.library_paths = []


def _diff(a, b):
    retval = ''
    started = False

    for group in difflib.SequenceMatcher(None, a, b).get_grouped_opcodes(3):
        if not started:
            started = True
            retval += '--- expected\n'
            retval += '+++ emitted\n'

        for tag, i1, i2, j1, j2 in group:
            if tag == 'equal':
                for line in a[i1:i2]:
                    for l in line.split('\n'):
                        retval += ' ' + l + '\n'
                continue

            if tag in ('replace', 'delete'):
                for line in a[i1:i2]:
                    for l in line.split('\n'):
                        retval += '-' + l + '\n'

            if tag in ('replace', 'insert'):
                for line in b[j1:j2]:
                    for l in line.split('\n'):
                        retval += '+' + l + '\n'

    return retval


def _extract_expected(filename):
    fd = open(filename, 'r', encoding='utf-8')
    data = fd.read()
    fd.close()

    retval = []
    for line in data.split('\n'):
        if line.startswith('// EXPECT:'):
            retval.append(line[10:] + '\n')
        elif line.startswith('//+'):
            retval[-1] += line[3:] + '\n'

    return retval


def check(args):
    filename = args[0]

    output = ChunkedIO()
    namespace = Namespace('Test', '1.0')
    logger = MessageLogger.get(namespace=namespace, output=output)
    logger.enable_warnings(True)
    logger.enable_strict(True)

    transformer = Transformer(namespace)
    transformer.set_include_paths([
        os.path.join(top_srcdir, 'gir'),
        top_builddir,
        os.path.join(top_builddir, 'gir'),
    ])
    transformer.register_include(Include.from_string('GObject-2.0'))

    ss = SourceScanner()

    options = Options()
    exit_code = process_packages(options, ['gobject-2.0'])
    if exit_code:
        sys.exit(exit_code)
    ss.set_cpp_options(options.cpp_includes, options.cpp_defines, options.cpp_undefines)
    ss.parse_files([filename])
    ss.parse_macros([filename])
    transformer.parse(ss.get_symbols())

    cbp = GtkDocCommentBlockParser()
    blocks = cbp.parse_comment_blocks(ss.get_comments())

    main = MainTransformer(transformer, blocks)
    main.transform()

    final = IntrospectablePass(transformer, blocks)
    final.validate()

    emitted_warnings = [w[w.find(':') + 1:] for w in output.getvalue()]

    expected_warnings = _extract_expected(filename)

    sortkey = lambda x: int(x.split(':')[0])
    expected_warnings.sort(key=sortkey)
    emitted_warnings.sort(key=sortkey)

    if len(expected_warnings) != len(emitted_warnings):
        raise SystemExit("ERROR in '%s': %d warnings were emitted, "
                         "expected %d:\n%s" % (os.path.basename(filename),
                                               len(emitted_warnings),
                                               len(expected_warnings),
                                               _diff(expected_warnings, emitted_warnings)))

    for emitted_warning, expected_warning in zip(emitted_warnings, expected_warnings):
        if expected_warning != emitted_warning:
            raise SystemExit("ERROR in '%s': expected warning does not match emitted "
                             "warning:\n%s" % (filename,
                                               _diff([expected_warning], [emitted_warning])))


sys.exit(check(sys.argv[1:]))



from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor


class RemoveOuterP(Treeprocessor):
    def run(self, root):
        if len(root) == 1 and root[0].tag == "p":
            root[0].tag = "span"


class InlineMarkdown(Extension):
    def extendMarkdown(self, md):
        md.treeprocessors.register(RemoveOuterP(md), "remove_outer_p", 0)



import unittest

try:
    import mako
except ImportError:
    HAS_MAKO = False
else:
    HAS_MAKO = True
    mako
    from giscanner.docwriter import DocWriter
    from giscanner.transformer import Transformer


@unittest.skipUnless(HAS_MAKO, "mako missing")
class TestDocWriter(unittest.TestCase):

    def test_main(self):
        transformer = Transformer(None)
        DocWriter(transformer, "gjs", "devdocs")


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python -*-
# GObject-Introspection - a framework for introspecting GObject libraries
# Copyright (C) 2008  Johan Dahlin
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

import os
import tempfile

from .message import Position
from .ccompiler import CCompiler
from .utils import have_debug_flag, dll_dirs


dlldirs = dll_dirs()
dlldirs.add_dll_dirs(['gio-2.0'])
from giscanner._giscanner import SourceScanner as CSourceScanner
dlldirs.cleanup_dll_dirs()

HEADER_EXTS = ['.h', '.hpp', '.hxx']
SOURCE_EXTS = ['.c', '.cpp', '.cc', '.cxx']
ALL_EXTS = SOURCE_EXTS + HEADER_EXTS

(CSYMBOL_TYPE_INVALID,
 CSYMBOL_TYPE_ELLIPSIS,
 CSYMBOL_TYPE_CONST,
 CSYMBOL_TYPE_OBJECT,
 CSYMBOL_TYPE_FUNCTION,
 CSYMBOL_TYPE_FUNCTION_MACRO,
 CSYMBOL_TYPE_STRUCT,
 CSYMBOL_TYPE_UNION,
 CSYMBOL_TYPE_ENUM,
 CSYMBOL_TYPE_TYPEDEF,
 CSYMBOL_TYPE_MEMBER) = range(11)

(CTYPE_INVALID,
 CTYPE_VOID,
 CTYPE_BASIC_TYPE,
 CTYPE_TYPEDEF,
 CTYPE_STRUCT,
 CTYPE_UNION,
 CTYPE_ENUM,
 CTYPE_POINTER,
 CTYPE_ARRAY,
 CTYPE_FUNCTION) = range(10)

STORAGE_CLASS_NONE = 0
STORAGE_CLASS_TYPEDEF = 1 << 1
STORAGE_CLASS_EXTERN = 1 << 2
STORAGE_CLASS_STATIC = 1 << 3
STORAGE_CLASS_AUTO = 1 << 4
STORAGE_CLASS_REGISTER = 1 << 5
STORAGE_CLASS_THREAD_LOCAL = 1 << 6

TYPE_QUALIFIER_NONE = 0
TYPE_QUALIFIER_CONST = 1 << 1
TYPE_QUALIFIER_RESTRICT = 1 << 2
TYPE_QUALIFIER_VOLATILE = 1 << 3
TYPE_QUALIFIER_EXTENSION = 1 << 4

FUNCTION_NONE = 0
FUNCTION_INLINE = 1 << 1

(UNARY_ADDRESS_OF,
 UNARY_POINTER_INDIRECTION,
 UNARY_PLUS,
 UNARY_MINUS,
 UNARY_BITWISE_COMPLEMENT,
 UNARY_LOGICAL_NEGATION) = range(6)


def symbol_type_name(symbol_type):
    return {
        CSYMBOL_TYPE_INVALID: 'invalid',
        CSYMBOL_TYPE_ELLIPSIS: 'ellipsis',
        CSYMBOL_TYPE_CONST: 'const',
        CSYMBOL_TYPE_OBJECT: 'object',
        CSYMBOL_TYPE_FUNCTION: 'function',
        CSYMBOL_TYPE_FUNCTION_MACRO: 'function_macro',
        CSYMBOL_TYPE_STRUCT: 'struct',
        CSYMBOL_TYPE_UNION: 'union',
        CSYMBOL_TYPE_ENUM: 'enum',
        CSYMBOL_TYPE_TYPEDEF: 'typedef',
        CSYMBOL_TYPE_MEMBER: 'member'}.get(symbol_type)


def ctype_name(ctype):
    return {
        CTYPE_INVALID: 'invalid',
        CTYPE_VOID: 'void',
        CTYPE_BASIC_TYPE: 'basic',
        CTYPE_TYPEDEF: 'typedef',
        CTYPE_STRUCT: 'struct',
        CTYPE_UNION: 'union',
        CTYPE_ENUM: 'enum',
        CTYPE_POINTER: 'pointer',
        CTYPE_ARRAY: 'array',
        CTYPE_FUNCTION: 'function'}.get(ctype)


class SourceType(object):
    __members__ = ['type', 'base_type', 'name', 'type_qualifier',
                   'child_list', 'is_bitfield', 'function_specifier']

    def __init__(self, scanner, stype):
        self._scanner = scanner
        self._stype = stype

    def __repr__(self):
        return "<%s type='%s' name='%s'>" % (
            self.__class__.__name__,
            ctype_name(self.type),
            self.name)

    @property
    def type(self):
        return self._stype.type

    @property
    def base_type(self):
        if self._stype.base_type is not None:
            return SourceType(self._scanner, self._stype.base_type)

    @property
    def name(self):
        return self._stype.name

    @property
    def type_qualifier(self):
        return self._stype.type_qualifier

    @property
    def child_list(self):
        for symbol in self._stype.child_list:
            if symbol is None:
                continue
            yield SourceSymbol(self._scanner, symbol)

    @property
    def is_bitfield(self):
        return self._stype.is_bitfield

    @property
    def function_specifier(self):
        return self._stype.function_specifier


class SourceSymbol(object):
    __members__ = ['const_int', 'const_double', 'const_string', 'const_boolean',
                   'ident', 'type', 'base_type']

    def __init__(self, scanner, symbol):
        self._scanner = scanner
        self._symbol = symbol

    def __repr__(self):
        src = self.source_filename
        if src:
            line = self.line
            if line:
                src += ":'%s'" % (line, )
        return "<%s type='%s' ident='%s' src='%s'>" % (
            self.__class__.__name__,
            symbol_type_name(self.type),
            self.ident,
            src)

    @property
    def const_int(self):
        return self._symbol.const_int

    @property
    def const_double(self):
        return self._symbol.const_double

    @property
    def const_string(self):
        return self._symbol.const_string

    @property
    def const_boolean(self):
        return self._symbol.const_boolean

    @property
    def ident(self):
        return self._symbol.ident

    @property
    def type(self):
        return self._symbol.type

    @property
    def base_type(self):
        if self._symbol.base_type is not None:
            return SourceType(self._scanner, self._symbol.base_type)

    @property
    def source_filename(self):
        return self._symbol.source_filename

    @property
    def line(self):
        return self._symbol.line

    @property
    def private(self):
        return self._symbol.private

    @property
    def position(self):
        return Position(self._symbol.source_filename,
                        self._symbol.line)


class SourceScanner(object):

    def __init__(self):
        self._scanner = CSourceScanner()
        self._filenames = []
        self._cpp_options = []
        self._compiler = None

    # Public API

    def set_cpp_options(self, includes, defines, undefines, cflags=[]):
        self._cpp_options.extend(cflags)
        for prefix, args in [('-I', [os.path.realpath(f) for f in includes]),
                             ('-D', defines),
                             ('-U', undefines)]:
            for arg in (args or []):
                opt = prefix + arg
                if opt not in self._cpp_options:
                    self._cpp_options.append(opt)

    def set_compiler(self, compiler):
        self._compiler = compiler

    def parse_files(self, filenames):
        for filename in filenames:
            # self._scanner expects file names to be canonicalized and symlinks to be resolved
            filename = os.path.realpath(filename)
            self._scanner.append_filename(filename)
            self._filenames.append(filename)

        headers = []
        for filename in self._filenames:
            if os.path.splitext(filename)[1] in SOURCE_EXTS:
                self._scanner.lex_filename(filename)
            else:
                headers.append(filename)

        self._parse(headers)

    def parse_macros(self, filenames):
        self._scanner.set_macro_scan(True)
        # self._scanner expects file names to be canonicalized and symlinks to be resolved
        self._scanner.parse_macros([os.path.realpath(f) for f in filenames])
        self._scanner.set_macro_scan(False)

    def get_symbols(self):
        for symbol in self._scanner.get_symbols():
            yield SourceSymbol(self._scanner, symbol)

    def get_comments(self):
        return self._scanner.get_comments()

    def get_errors(self):
        return self._scanner.get_errors()

    def dump(self):
        print('-' * 30)
        for symbol in self._scanner.get_symbols():
            print(symbol.ident, symbol.base_type.name, symbol.type)

    # Private

    def _parse(self, filenames):
        if not filenames:
            return

        defines = ['__GI_SCANNER__']
        undefs = []

        cc = CCompiler(compiler_name=self._compiler)

        tmp_fd_cpp, tmp_name_cpp = tempfile.mkstemp(prefix='g-ir-cpp-',
                                                    suffix='.c',
                                                    dir=os.getcwd())
        with os.fdopen(tmp_fd_cpp, 'wb') as fp_cpp:
            self._write_preprocess_src(fp_cpp, defines, undefs, filenames)

        tmpfile_basename = os.path.basename(os.path.splitext(tmp_name_cpp)[0])

        # Output file name of the preprocessor, only really used on non-MSVC,
        # so we want the name to match the output file name of the MSVC preprocessor
        tmpfile_output = tmpfile_basename + '.i'

        cc.preprocess(tmp_name_cpp,
                      tmpfile_output,
                      self._cpp_options)

        if not have_debug_flag('save-temps'):
            os.unlink(tmp_name_cpp)
        self._scanner.parse_file(tmpfile_output)
        if not have_debug_flag('save-temps'):
            os.unlink(tmpfile_output)

    def _write_preprocess_src(self, fp, defines, undefs, filenames):
        # Write to the temp file for feeding into the preprocessor
        for define in defines:
            fp.write(('#ifndef %s\n' % (define, )).encode())
            fp.write(('# define %s\n' % (define, )).encode())
            fp.write('#endif\n'.encode())
        for undef in undefs:
            fp.write(('#undef %s\n' % (undef, )).encode())
        for filename in filenames:
            fp.write(('#include <%s>\n' % (filename, )).encode())
